/** 
 * @file hdc1000_click_example.js
 * @brief Module for Mikro Elektronika HDC1000 Click board.
 * 
 * @note: Start up notes
 * This is an example of how to use HDC1000 clik board on your beaglebone. Module 
 * exports main function and global variables but if u want to test it - uncomment 
 * line 169:runtest() and run it from command line with command
 * $ node hdc1000_click_example.js
 */
 
var command = require('./commands-hdc1000')();                                  /**< module with commands for MikroE hdc1000 click */
var i2c = require('i2c');                                                       /**< module for i2c communication */

var address = 0x40;                                                             /**< hdc1000 hardware address */
var hdc1000 = new i2c(address, {device: '/dev/i2c-1'});                         /**< i2c bus initialisation */

var t = new Buffer(2);
var h = new Buffer(2);
var tempConst = 0.0025177;                                                      /**< Temperature constant from datasheet */
var humConst = 0.0015259;                                                       /**< Humidity constant from datasheet */
var temperature, humidity, dewPoint;                                            /**< used for storing current weather conditions */

/**
 * @brief HDC1000 initialisation
 * 
 * Write initialisation command to device
 */
function initDevice(callback) {
    hdc1000.write(command["Initialisation"], function(err) {
        if (err) {
            console.log(err);
        }
    });
    setTimeout(callback,20);
}

/**
 * @brief Handle synchronous execution of next function
 */
function handleTempCommand(){
    tempCommand(handleReadTemp);
}

/**
 * @brief Writes temperature command to HDC1000
 * 
 * Time delay 20ms is needed for AD conversion.
 */
function tempCommand(callback) {
    hdc1000.write(command["ReadTemperature"], function(err) {
        if (err) {
            console.log(err)
        }
    });
    setTimeout(callback,20);
}

/**
 * @brief Handle synchronous execution of next function
 */
function handleReadTemp(){
    readTemp(handleHumCommand);
}

/**
 * @brief Reads temperature from HDC1000
 * 
 * Temperature value in raw form is stored into 16 bits long buffer (T).
 */
function readTemp(callback) {
    hdc1000.read(2, function(err, res) {
        if (err) {
            console.log(err)
        }
        else {
            t[0] = res[0];
            t[1] = res[1];
        }
    });
    setTimeout(callback);
}

/**
 * @brief Handle synchronous execution of next function
 */
function handleHumCommand(){
    humCommand(handleReadHum);
}

/**
 * @brief Writes humidity command to HDC1000
 * 
 * Time delay 20ms is needed for AD conversion.
 */
function humCommand(callback){
    hdc1000.write(command["ReadHumidity"], function(err) {
        if (err) {
            console.log(err)
        }
    });
    setTimeout(callback,20);
}

/**
 * @brief Handle synchronous execution of next functiontion
 */
function handleReadHum(){
    readHum(handleCalculate);
}

/**
 * @brief Reads humididty from HDC1000
 * 
 * Humidity value in raw form is stored into 16 bits long buffer (H).
 */
function readHum(callback){
    hdc1000.read(2, function(err, res) {
        if (err) {
            console.log(err)
        }
        else {
            h[0] = res[0];
            h[1] = res[1];
        }
    });
    setTimeout(callback,10);
}

/**
 * @brief Handle synchronous execution of next functiontion
 */
function handleCalculate(){
    calculate();
}

/**
 * @brief Calculates temperature and humidity values
 * 
 * Converts buffer to int value and calculate temperature, humidity and dew point
 * Displays all values in console.
 * 
 */
function calculate() {
    var temp1 = t.readInt16BE(0);
    var temp2 = h.readInt16BE(0);
    
    temperature = ( temp1 * tempConst ) - 40;
    humidity = temp2 * humConst;
    
    dewPoint = 243.04 *(Math.log(humidity/100) +((17.625 * temperature)/(243.04 + temperature)))
                /(17.625 - Math.log(humidity/100)-((17.625*temperature)/(243.04 + temperature)));
    
    console.log('Temperature : ' + temperature.toFixed(1) + ' C');
    console.log('Humidity : ' + humidity.toFixed(0) + ' %');
    console.log('Dew point : ' + dewPoint.toFixed(1) + ' C');
}

/**
 * @brief Main function
 * 
 * Mesurement of current conditions
 */
function readSensor(){
    initDevice(handleTempCommand);
}

//readSensor();

/**
 * @brief Used for exports temperature value
 */
function temperature(){
    return temperature;
}

/**
 * @brief Used for exports humidity value
 */
function humidity(){
    return humidity;
}

/**
 * @brief Used for exports dew point value
 */
function dewPoint(){
    return dewPoint;
}


module.exports = {
    measure: function(){
        readSensor()
    },
    temperature: function(){
        return temperature;
    },
    humidity: function(){
        return humidity;
    },
    dewPoint: function(){
        return dewPoint;
    }
}