/**
 * @file pressure2_clik_example.js
 * @brief Module for Mikro Elektronika Pressure2 Click board.
 * 
 * @note: Start up notes
 * This is an example of how to use Pressure2 click board on your beaglebone.
 * Module exports main function and global variables but if u want to test it -
 * uncomment line:227 readSensor() and run it from command line with command 
 * $ node pressure2_click_example.js
 * 
 * If needed, edit SPI bus configuration depending on mikroBUS socket
 *          - mikroBUS Cape socket 1    /dev/spidev1.0
 *          - mikroBUS Cape socket 2    /dev/spidev1.1
 *          - mikroBUS Cape socket 3    /dev/spidev2.0
 *          - mikroBUS Cape socket 4    /dev/spidev1.2
 */

var command = require('./commands-pressure2')();                                /**< module with commands for MikroE Pressure2 Click */
var SPI = require('spi-trp');                                                   /**< module for spi comunication  */

/**
 * @brief SPI bus configuration
 */
var spi = new SPI.Spi('/dev/spidev1.0',{
    'mode' : SPI.MODE['MODE_3']}                                                /**< mode setup - 'MODE_0'/'MODE_1'/'MODE_2'/'MODE_3' */
    );

var sensorCoef = [];                                                            /**< used for storing sensor coeficiens */
var d1, d2, t, p;                                                               /**< used for temperature and pressure */

/**
 * @brief Opens SPI bus and reset device
 */
function openDevice(callback){
    
    spi.open();
    spi.write(command["Reset"]);
    
    callback();
}

/**
 * @brief Handle synchronous execution of next function
 */
function handleReadCoefs(){
    readCoefs(handleConvertD1);
}

/**
 * @brief Read sensor coeficients 
 * 
 * There is eight sensor coeficients and all of them are needed for temperature
 * and pressure values calculation. Functions reads all of them and store values
 * into array.
 */
function readCoefs(callback){
    
    for (var i = 0; i < 8; i++)
    {
        var buf = new Buffer(3);
        
        spi.transfer(command[i],buf, function(){
        var realBuf = new Buffer(2);
        
        realBuf[0] = buf[1];
        realBuf[1] = buf[2];
        sensorCoef[i] = realBuf.readUInt16BE(0);
        });
    }
    
    callback();
}

/**
 * @brief Handle synchronous execution of next function
 */
function handleConvertD1(){
    convertD1(handleReadD1);
}

/**
 * @brief Writes D1 conversion command
 * 
 * Time delay needed for AD conversion.
 */
function convertD1(callback){
    
    spi.write(command["ConvertD1"]);
    setTimeout(callback,14);
}

/**
 * @brief Handle synchronous execution of next function
 */
function handleReadD1(){
    readD1(handleConvertD2);
}

/**
 * @brief Read D1 value from sensor.
 */
function readD1(callback){
    var buf = new Buffer(4);
    
    spi.transfer(command["ADC-Read"], buf, function(){                          
    var realBuf = new Buffer(4);
    
    realBuf[0] = [0x00];
    realBuf[1] = buf[1];
    realBuf[2] = buf[2];
    realBuf[3] = buf[3];
    d1 = realBuf.readUInt32BE(0);                    
    });
    
    callback();
}

/**
 * @brief Handle synchronous execution of next function
 */
function handleConvertD2(){
    convertD2(handleReadD2);
}

/**
 * @brief Writes D2 conversion command
 * 
 * Time delay needed for AD conversion.
 */
function convertD2(callback){
    
    spi.write(command["ConvertD2"]);
    setTimeout(callback,14);
}

/**
 * @brief Handle synchronous execution of next function
 */
function handleReadD2(){
    readD2(handleCalculateValues);
}

/**
 * @brief Read D2 value from sensor.
 */
function readD2(callback){
    
    var buf = new Buffer(4);
    
    spi.transfer(command["ADC-Read"], buf, function(){
    var realBuf = new Buffer(4);
    
    realBuf[0] = [0x00];
    realBuf[1] = buf[1];
    realBuf[2] = buf[2];
    realBuf[3] = buf[3];
    d2 = realBuf.readUInt32BE(0);
    });
    
    callback();
}

/**
 * @brief Handle synchronous execution of next function
 */
function handleCalculateValues(){
    calculateValues(handleDisplayValues);
}

/**
 * @brief Calculates pressure and temperature
 */
function calculateValues(callback){
    
    var deltaTemp = d2 - sensorCoef[5] * Math.pow(2,8);
    var sensorOffset = sensorCoef[2] * Math.pow(2,16) + ( deltaTemp*sensorCoef[4] ) / Math.pow(2,7);
    var sensitivity = sensorCoef[1] * Math.pow(2,15) + ( deltaTemp*sensorCoef[3] ) / Math.pow(2,8);
    
    p = ((( d1 * sensitivity ) / Math.pow(2,21) - sensorOffset ) / Math.pow(2,15)) / 10 ;
    t = ( 2000 + (deltaTemp * sensorCoef[6]) / Math.pow(2,23)) / 100;
    
    callback();
}

/**
 * @brief Handle synchronous execution of next function
 */
function handleDisplayValues(){
    displayValues(handleCloseDevice);
}

/**
 * @brief Displays values on console
 */
function displayValues(callback){
    
    console.log("Current Temperature     :     ",t.toFixed(1) + " C");
    console.log("Current Preassure       :     ",p.toFixed(0) + " mB");
    
    callback();
}

/**
 * @brief Handle synchronous execution of next function
 */
function handleCloseDevice(){
    closeDevice();
}

/**
 * Closes SPI bus
 */
function closeDevice(){
    
    spi.close();
}

/**
 * @brief Main function - measuremet of current conditions
 * 
 * Executes all synchronous.
 */
function readSensor(){
    openDevice(handleReadCoefs);
}

//readSensor();

/**
 * @brief Used for exports temperature value
 */
function temperature(){
    return t;
}

/**
 * @brief Used for exports pressure value
 */
function pressure(){
    return p;
}


module.exports = {
    measure: function(){
        readSensor();
    },
    temperature: function(){
        return t;
    },
    pressure: function(){
        return p;
    }
}